import 'dart:io';
import 'dart:async';
import 'package:adeptarch/adeptarch.dart';
import 'package:adeptarch/src/api/adept_graphql_client.dart';
import 'package:adeptutils/adeptutils.dart' as AdeptUtils;
import 'package:flutter/foundation.dart';
import 'package:graphql_flutter/graphql_flutter.dart';
import 'package:http/http.dart' as http;

GraphQLAPI graphQLAPI = GraphQLAPI();

class GraphQLAPI
{
  AdeptGraphQLClient client;
  ValueNotifier<AdeptGraphQLClient> clientNotifier;
  http.Client httpClient;
  Cache cache;
  Link apiLink;
  String apiUrl;
  String wsApiUrl;

  Duration timeout;
  GraphQLAPI(){
    clientNotifier = ValueNotifier(client);
  }

  void init({Link link, String apiUrl, String wsApiUrl, GetToken getToken, Duration timeout, http.Client httpClient})
  {
    if(apiUrl != null) this.apiUrl = apiUrl;
    if(wsApiUrl != null) this.wsApiUrl = wsApiUrl;
    if(timeout != null) this.timeout = timeout;
    if(httpClient != null) this.httpClient = httpClient;
    if(link != null) apiLink = link; else {
      AuthLink authLink = getToken != null ? AuthLink(getToken: getToken) : null;
      HttpLink httpLink = HttpLink(uri: this.apiUrl, httpClient: this.httpClient);
      WebSocketLink webSocketLink = this.wsApiUrl != null && this.wsApiUrl.isNotEmpty ? WebSocketLink(url: this.wsApiUrl,) : null;
      apiLink = null;
      if(authLink != null) apiLink = authLink;
      if(httpLink != null){
        if(apiLink != null) apiLink = apiLink.concat(httpLink); else
          apiLink = httpLink;
      }
      if(webSocketLink != null){
        if(apiLink != null) apiLink = apiLink.concat(webSocketLink); else
          apiLink = webSocketLink;
      }
    }

    client = AdeptGraphQLClient(
      link: apiLink,
      cache: cache ?? InMemoryCache(),
      timeout: timeout,
    );
    clientNotifier.value = client;
  }

  ///Implement this method when checked what exceptions could be
  String getLocalizedMessageFromException(dynamic exception)
  {
    if(exception == null) return null;
    if(exception is AdeptUtils.GenericException) return getLocalizedMessageFromException(exception.causeException);

    if(exception is WebSocketException || exception is TimeoutException ||
        exception.toString().toLowerCase().contains("TimeoutException"))
      return "Server connection timeout.";
    if(exception is SocketException || exception.toString().toLowerCase().contains("failed to connect"))
      return "Check your internet connection. ${exception.toString()}";
    if(exception is HandshakeException)
      return exception.toString();

    return exception.toString();
  }

  String getGenericLocalizedMessageFromException(dynamic exception)
  {
    String message = getLocalizedMessageFromException(exception);
    return message ?? "Unknown error";
  }

  String getLocalizedMessageFromHttpResultCode(int code)
  {
    if(code == null) return null;
    switch(code)
    {
      case HttpStatus.noContent:
        return "($code) No Content";

      case HttpStatus.badRequest:
        return "($code) Bad Request";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized";

      case HttpStatus.forbidden:
        return "($code) Forbidden";

      case HttpStatus.notFound:
        return "($code) Not Found";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error";
    }
    return "HTTP error code: ($code)";
  }

  String getLocalizedMessageFromResponse({QueryResponse response, int code})
  {
    code ??= response?.errorStatusCode;
    if(code == null) return null;
    String extraMessage = " ${response?.errorMessage != null ? response?.errorMessage : ""}";
    switch(code)
    {
      case HttpStatus.noContent:
        return "($code) No Content.$extraMessage";

      case HttpStatus.badRequest:
        return "($code) Bad Request.$extraMessage";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized.$extraMessage";

      case HttpStatus.forbidden:
        return "($code) Forbidden.$extraMessage";

      case HttpStatus.notFound:
        return "($code) Not Found.$extraMessage";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error.$extraMessage";

      case AdeptUtils.Jsonable.jsonParserError:
        return "($code) Invalid response. JsonParser was unable to decode response.$extraMessage";
    }
    return "HTTP error code: ($code).$extraMessage";
  }
}