import 'package:adeptarch/adeptarch.dart';
import 'package:adeptarch/src/api/graphql_base_error.dart';
import 'package:adeptutils/adeptutils.dart';
import 'package:graphql_flutter/graphql_flutter.dart';

class QueryResponse<DataType, ErrorType> {

  GraphQLError _mainGraphQLError;
  DateTime timestamp;
  List<ErrorType> _errors;
  ErrorType _mainError;
  ErrorType _errorType;

  /// The source of the result data.
  ///
  /// null when unexecuted.
  /// Will be set when encountering an error during any execution attempt
  QueryResultSource source;

  DataType data;

  OperationException exception;

  QueryResponse({
    this.data,
    ErrorType errorType,
    this.exception,
    bool loading,
    bool optimistic,
    QueryResultSource source,
    DateTime timestamp,
  }) :
    _errorType = errorType ?? GraphQLBaseError(),
    timestamp = timestamp ?? DateTime.now(),
    this.source = source ??
      ((loading == true)
        ? QueryResultSource.Loading
        : (optimistic == true)
        ? QueryResultSource.OptimisticResult
        : null);

  List<ErrorType> get errors {
    if(_errors == null && hasException) {
      if((exception?.graphqlErrors?.isNotEmpty ?? false)){
        _errors = [];
        try{
          exception.graphqlErrors.forEach((graphQLError) {
            if(_errorType is Jsonable)
              _errors.add((_errorType as Jsonable).fromJsonMap(graphQLError.raw));
            else _errors.add(graphQLError.raw as ErrorType);
          });
        }catch(e){
          print(e);
        }
      }

    }
    return _errors;
  }

  ErrorType get mainError {
    if(_mainError == null && (errors?.isNotEmpty ?? false)) _mainError = errors[0];
    return _mainError;
  }

  GraphQLError get mainGraphQLError {
    if(_mainGraphQLError == null){
      if((exception?.graphqlErrors?.isNotEmpty ?? false))
        _mainGraphQLError = exception?.graphqlErrors[0];
    }
    return _mainGraphQLError;
  }

  String get errorMessage {
    try{
      if(mainGraphQLError != null){
        return mainGraphQLError?.message;
      }
    }catch(e){

    }
    return null;
  }

  String get errorCode {
    try{
      if(mainGraphQLError != null && (mainGraphQLError?.extensions?.isNotEmpty ?? false)){
        return mainGraphQLError.extensions['code'];
      }
    }catch(e){

    }
    return null;
  }

  int get errorStatusCode {
    try{
      if(mainGraphQLError != null && (mainGraphQLError?.extensions?.isNotEmpty ?? false)){
        Map<String, dynamic> exception = mainGraphQLError.extensions['exception'];
        if((exception?.isNotEmpty ?? false))
          return exception['status'];
      }
    }catch(e){

    }
    return null;
  }

  /// Whether data has been specified from either the cache or network)
  bool get loading => source == QueryResultSource.Loading;

  /// Whether an optimistic result has been specified
  ///   (may include eager results from the cache)
  bool get optimistic => source == QueryResultSource.OptimisticResult;

  /// Whether the response includes an exception
  bool get hasException => exception != null;

  /// Whether the response includes an exception
  bool get isSuccessful => exception == null;

//  static QueryResponse fromQueryResult(
//      QueryResult queryResult, DataTypeResponseConverter dataTypeResponseConverter){
//    var data;
//    try{
//      data = dataTypeResponseConverter(queryResult);
//    }catch(e){
//      print(e);
//    }
//    return QueryResponse(
//      data: data,
//      exception: queryResult?.exception,
//      source: queryResult?.source,
//      timestamp: queryResult.timestamp,
//    );
//  }
//
//  static QueryResponse fromQueryResultOld<DataType extends Jsonable>(
//      QueryResult queryResult, DataType dataType,
//      DataTypeResponseConverter dataTypeResponseConverter){
//    DataType data;
//    try{
//      if(queryResult?.data != null)
//        return queryResult?.data is Map
//            ? fromMapQueryResult(queryResult, dataType, dataTypeResponseConverter)
//            : fromListQueryResult(queryResult, dataType, dataTypeResponseConverter);
//    }catch(e){
//      print(e);
//    }
//    return QueryResponse<DataType>(
//      data: data,
//      exception: queryResult?.exception,
//      source: queryResult?.source,
//      timestamp: queryResult.timestamp,
//    );
//  }

  static QueryResponse<DataType, ErrorType>
    fromMapQueryResult<DataType, ErrorType>(
      QueryResult queryResult,
      ErrorType errorType,
      DataTypeQueryResponseConverter<DataType> dataTypeResponseConverter){
    DataType data;
    try{
      data = dataTypeResponseConverter(queryResult);
    }catch(e){
      print(e);
    }
    return QueryResponse<DataType, ErrorType>(
      data: data,
      errorType: errorType,
      exception: queryResult?.exception,
      source: queryResult?.source,
      timestamp: queryResult?.timestamp,
    );
  }

  static QueryResponse<List<DataType>, ErrorType>
    fromListQueryResult<DataType, ErrorType>(
      QueryResult queryResult,
      ErrorType errorType,
      DataTypeQueryResponseConverter<List<DataType>> dataTypeResponseConverter){
    List<DataType> data;
    try{
      data = dataTypeResponseConverter(queryResult);
    }catch(e){
      print(e);
    }
    return QueryResponse<List<DataType>, ErrorType>(
      data: data,
      errorType: errorType,
      exception: queryResult?.exception,
      source: queryResult?.source,
      timestamp: queryResult?.timestamp,
    );
  }
}