import 'dart:io';
import 'dart:async';
import 'package:adeptutils/adeptutils.dart' as AdeptUtils;
import 'package:chopper/chopper.dart';
import 'package:adeptarch/src/api/adept_chopper_client.dart';
import 'package:http/http.dart' as http;

RestAPI restAPI = RestAPI();

class RestAPI
{
  http.Client httpClient;
  AdeptChopperClient chopperClient;
  Converter converter;
  final Map<Type, ChopperService> _services = {};

  Duration timeout;
  String apiProtocol;
  String apiHost;
  String apiPort;
  String apiUrl;

  String webProtocol;
  String webHost;
  String webPort;
  String webUrl;

  String apiDateFormat;
  bool trustAllCerts;

  RestAPI(){
    apiProtocol = "http://";
    apiHost = "www.myapi.com";
    apiPort = "8080";
    apiUrl = apiProtocol + apiHost + apiPort ?? "" + "/api";
    apiDateFormat = "yyyy-MM-dd";

    webProtocol = "http://";
    webHost = "www.myweb.com";
    webPort = "";
    webUrl = webProtocol + webHost + webPort ?? "";

    trustAllCerts = true;
  }

  void init({String apiUrl, Duration timeout, http.Client httpClient})
  {
    if(AdeptUtils.TextUtils.isNotEmpty(apiUrl)) updateApiUrl(apiUrl: apiUrl);
    if(timeout != null) this.timeout = timeout;
    if(httpClient != null) this.httpClient = httpClient;
    try{chopperClient?.dispose();} catch(exception){}
    converter ??= JsonConverter();
    chopperClient = AdeptChopperClient(
      baseUrl: this.apiUrl,
      client: this.httpClient,
      services: _services.values,
      converter: this.converter,
      timeout: this.timeout,
    );
    _initServices();
  }

  void _initServices(){
    _services.forEach((type, service) {
      service.client = chopperClient;
    });
  }

  void dispose(){
    try{
      _services.forEach((t, s) => s.dispose());
      _services.clear();
      chopperClient.dispose();
    }
    catch(exception){}
  }

  void addService(ChopperService service){
    if(service != null) {
      service.client = chopperClient;
      _services[service.definitionType] = service;
    }
  }

  T service<T extends ChopperService>(Type type) {
    final s = _services[type];
    if (s == null) {
      throw Exception("Service of type '$type' not found.");
    }
    return s;
  }

  void updateApiUrl({String apiUrl, String apiProtocol, String apiHost,
    String apiPort})
  {
    if(apiUrl == null || apiUrl.isEmpty){
      this.apiProtocol = apiProtocol;
      this.apiHost = apiHost;
      this.apiPort = apiPort;
      this.apiUrl = apiProtocol + apiHost + apiPort ?? "" + "/api/";
    } else{
      this.apiUrl = apiUrl;
    }
  }

  void updateWebUrl({String webUrl, String webProtocol, String webHost,
    String webPort})
  {
    if(webUrl == null || webUrl.isEmpty){
      this.webProtocol = webProtocol;
      this.webHost = webHost;
      this.webPort = webPort;
      this.webUrl = webProtocol + webHost + webPort ?? "";
    } else {
      this.webUrl = webUrl;
    }
  }
  
  ///Implement this method when checked what exceptions could be
  String getLocalizedMessageFromException(dynamic exception)
  {
    if(exception == null) return null;
    if(exception is AdeptUtils.GenericException) return getLocalizedMessageFromException(exception.causeException);

    if(exception is WebSocketException || exception is TimeoutException ||
        exception.toString().toLowerCase().contains("TimeoutException"))
      return "Server connection timeout.";
    if(exception is SocketException || exception.toString().toLowerCase().contains("failed to connect"))
      return "Check your internet connection. ${exception.toString()}";
    if(exception is HandshakeException)
      return exception.toString();

    return exception.toString();
  }

  String getGenericLocalizedMessageFromException(dynamic exception)
  {
    String message = getLocalizedMessageFromException(exception);
    return message ?? "Unknown error";
  }

  String getLocalizedMessageFromHttpResultCode(int code)
  {
    if(code == null) return null;
    switch(code)
    {
      case HttpStatus.noContent:
        return "($code) No Content";

      case HttpStatus.badRequest:
        return "($code) Bad Request";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized";

      case HttpStatus.forbidden:
        return "($code) Forbidden";

      case HttpStatus.notFound:
        return "($code) Not Found";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error";
    }
    return "HTTP error code: ($code)";
  }

  String getLocalizedMessageFromResponse({Response response, int code})
  {
    code ??= response?.statusCode;
    if(code == null) return null;
    String extraMessage = " ${response?.error != null ? response?.error : ""}";
    switch(code)
    {
      case HttpStatus.noContent:
        return "($code) No Content.$extraMessage";

      case HttpStatus.badRequest:
        return "($code) Bad Request.$extraMessage";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized.$extraMessage";

      case HttpStatus.forbidden:
        return "($code) Forbidden.$extraMessage";

      case HttpStatus.notFound:
        return "($code) Not Found.$extraMessage";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error.$extraMessage";

      case AdeptUtils.Jsonable.jsonParserError:
        return "($code) Invalid response. JsonParser was unable to decode response.$extraMessage";
    }
    return "HTTP error code: ($code).$extraMessage";
  }
}