import 'package:adeptarch/src/filters/filter.dart';
import 'package:adeptarch/src/utils/sort_mode.dart';
import 'package:adeptutils/adeptutils.dart';
import 'package:json_annotation/json_annotation.dart';
 import 'package:flutter/foundation.dart';

part 'date_filter.g.dart';

const int
    DATE_FILTER_ALL = 0,
    DATE_FILTER_YESTERDAY = 1,
    DATE_FILTER_TODAY = 2,
    DATE_FILTER_NOW = 3,
    DATE_FILTER_TOMORROW = 4,
    DATE_FILTER_BEFORE_TODAY = 5,
    DATE_FILTER_AFTER_TODAY = 6,
    DATE_FILTER_RANGE = 7
;

@JsonSerializable()
class DateFilter extends Filter<DateFilter, String>
{
  bool dateFilterAsTimeMillis;
  int dateFilterType;
  String _paramDateStartGreaterOrEqual,
      _paramDateEndLessOrEqual;

  String _paramStart, _paramEnd;

  ZonedDateTime dateStart, dateEnd;

  DateFilter(
    {
      String param,
      @required String paramStart,
      @required String paramEnd,
      SortMode sortMode,
      FilterActionType filterActionType,
      this.dateFilterType = DATE_FILTER_ALL,
      this.dateFilterAsTimeMillis,
    })
      : super(
      param: param ?? paramStart,
      sortMode: sortMode,
      filterActionType: filterActionType,
    )
  {
    this.dateFilterAsTimeMillis ??= false;
    this.paramStart = paramStart ?? param;
    this.paramEnd = paramEnd;
    _initCompareParams();
  }

  @override
  bool get isReady {
    if(param == null) return false;
    if(sortMode != null) return true;
    if(paramStart == null && paramEnd == null) return false;
    if(dateFilterType == DATE_FILTER_ALL) return false;
    return isValid;
  }

  @override
  bool get isValid {
    if(dateFilterType == DATE_FILTER_RANGE){
      if(dateStart == null && dateEnd == null) return false; else
      if(dateStart != null && dateEnd != null)
        if(dateStart.getDateTimeMillis() > dateEnd.getDateTimeMillis()) return false;
    }
    return true;
  }

  @override
  void reset() {
    dateStart = null;
    dateEnd = null;
    dateFilterType = DATE_FILTER_ALL;
    super.reset();
  }

  String get paramStart => _paramStart;
  set paramStart(String value) {
    if(value == null) return;
    _paramStart = value;
    if(paramEnd == null && value != null) paramEnd = value;
    _initCompareParams();
  }

  get paramEnd => _paramEnd;
  set paramEnd(value) {
    if(value == null) return;
    _paramEnd = value;
    if(paramStart == null && value != null) paramStart = value;
    _initCompareParams();
  }

  void _initCompareParams(){
    _paramDateStartGreaterOrEqual = "$_paramStart${FilterActions.GREATER_OR_EQUAL_THAN}";
    _paramDateEndLessOrEqual = "$_paramEnd${FilterActions.LESS_OR_EQUAL_THAN}";
  }

  void setDateStart({ZonedDateTime zonedDateTime, DateTime dateTime, int dateTimeMillis}){
    if(dateTimeMillis != null)
      return setDateStart(zonedDateTime: ZonedDateTime(dateTimeMillis: dateTimeMillis));
    if(dateTime != null)
      return setDateStart(zonedDateTime: ZonedDateTime(dateAsDateTime: dateTime));
    switch(dateFilterType){
      case DATE_FILTER_RANGE:
        dateStart = zonedDateTime;
        break;
    }
  }

  ZonedDateTime getDateStart(){
    switch(dateFilterType){
      case DATE_FILTER_ALL:
        return null;
      case DATE_FILTER_YESTERDAY:
        return new ZonedDateTime(dateTimeMillis: DateUtils.yesterday());
      case DATE_FILTER_TODAY:
        return new ZonedDateTime(dateTimeMillis: DateUtils.today());
      case DATE_FILTER_NOW:
        return new ZonedDateTime(dateTimeMillis: DateUtils.nowUntilMinutes());
      case DATE_FILTER_TOMORROW:
        return new ZonedDateTime(dateTimeMillis: DateUtils.tomorrow());
      case DATE_FILTER_BEFORE_TODAY:
        return null;
      case DATE_FILTER_AFTER_TODAY:
        return new ZonedDateTime(dateTimeMillis: DateUtils.tomorrow());
      case DATE_FILTER_RANGE:
        return dateStart;
    }
    return dateStart;
  }

  dynamic getDateStartDynamic() => dateFilterAsTimeMillis ? getDateStart().getDateTimeMillis() : getDateStart().getDate();

  void setDateEnd({ZonedDateTime zonedDateTime, DateTime dateTime, int dateTimeMillis}){
    if(dateTimeMillis != null)
      return setDateEnd(zonedDateTime: ZonedDateTime(dateTimeMillis: dateTimeMillis));
    if(dateTime != null)
      return setDateEnd(zonedDateTime: ZonedDateTime(dateAsDateTime: dateTime));
    switch(dateFilterType){
      case DATE_FILTER_RANGE:
        dateEnd = zonedDateTime;
        break;
    }
  }

  ZonedDateTime getDateEnd(){
    switch(dateFilterType){
      case DATE_FILTER_ALL:
        return null;
      case DATE_FILTER_YESTERDAY:
        return new ZonedDateTime(dateTimeMillis: DateUtils.today());
      case DATE_FILTER_TODAY:
        return new ZonedDateTime(dateTimeMillis: DateUtils.tomorrow());
      case DATE_FILTER_NOW:
        return new ZonedDateTime(dateTimeMillis: DateUtils.tomorrow());
      case DATE_FILTER_TOMORROW:
        return new ZonedDateTime(dateTimeMillis: DateUtils.dayAfterTomorrow());
      case DATE_FILTER_BEFORE_TODAY:
        return new ZonedDateTime(dateTimeMillis: DateUtils.today());
      case DATE_FILTER_AFTER_TODAY:
        return null;
      case DATE_FILTER_RANGE:
        return dateEnd;
    }
    return dateEnd;
  }

  dynamic getDateEndDynamic() => dateFilterAsTimeMillis ? getDateEnd().getDateTimeMillis() : getDateEnd().getDate();

  Map buildDateFilter()
  {
    switch(dateFilterType){
      case DATE_FILTER_BEFORE_TODAY:
        if(TextUtils.isNotEmpty(_paramDateEndLessOrEqual) && getDateEnd() != null)
          filtersBuilder.putAll(FilterActions.build(filterActionType, paramEnd, FilterActions.LESS_OR_EQUAL_THAN, getDateEndDynamic()));
        break;
      case DATE_FILTER_AFTER_TODAY:
        if(TextUtils.isNotEmpty(_paramDateStartGreaterOrEqual) && getDateStart() != null)
          filtersBuilder.putAll(FilterActions.build(filterActionType, paramStart, FilterActions.GREATER_OR_EQUAL_THAN, getDateStartDynamic()));
        break;
      default:
        if(TextUtils.isNotEmpty(_paramDateStartGreaterOrEqual) && getDateStart() != null)
          filtersBuilder.putAll(FilterActions.build(filterActionType, paramStart, FilterActions.GREATER_OR_EQUAL_THAN, getDateStartDynamic()));
        if(TextUtils.isNotEmpty(_paramDateEndLessOrEqual) && getDateEnd() != null)
          filtersBuilder.putAll(FilterActions.build(filterActionType, paramEnd, FilterActions.LESS_OR_EQUAL_THAN, getDateEndDynamic()));
        break;
    }

    return filtersBuilder.build();
  }

  @override
  Map<String, dynamic> build() {
    super.build();
    return buildDateFilter();
  }

  @override
  Map<String, dynamic> toJson() => _$DateFilterToJson(this);
  @override
  DateFilter fromJsonMap(Map<String, dynamic> json) => DateFilter.fromJson(json);
  factory DateFilter.fromJson(Map<String, dynamic> json) => json != null ? _$DateFilterFromJson(json) : null;
}