import 'package:json_annotation/json_annotation.dart';
import 'package:adeptarch/src/utils/sort_mode.dart';
import 'package:adeptutils/adeptutils.dart';
import 'package:flutter/foundation.dart';

export 'package:adeptutils/src/date_utils.dart';
export 'package:adeptutils/src/text_utils.dart';
export 'package:adeptarch/src/utils/sort_mode.dart';
export 'package:adeptarch/src/filters/text_filter.dart';
export 'package:adeptarch/src/filters/number_filter.dart';
export 'package:adeptarch/src/filters/bool_filter.dart';
export 'package:adeptarch/src/filters/date_filter.dart';
export 'package:adeptarch/src/filters/enum_filter.dart';

class FilterActions
{
  static const String CONTAINS = ".contains";
  static const String EQUALS = ".equals";
  static const String IN = ".in";
  static const String SPECIFIED = ".specified";
  static const String GREATER_OR_EQUAL_THAN = ".greaterOrEqualThan";
  static const String GREATER_THAN = ".greaterThan";
  static const String LESS_OR_EQUAL_THAN = ".lessOrEqualThan";
  static const String LESS_THAN = ".lessThan";

  static Map<String, dynamic> build(FilterActionType filterActionType, String param, String action, dynamic value){
    if(filterActionType == null) filterActionType = FilterActionType.FILTER_JPA;
    if(action == null) return MapBuilder<String, dynamic>(multiKey: true).put(param, value).build();

    switch(filterActionType) {
      case FilterActionType.FILTER_NONE:
        return MapBuilder<String, dynamic>(multiKey: true).put(param, value).build();
      case FilterActionType.FILTER_JPA:
        return MapBuilder<String, dynamic>(multiKey: true).put(param + action, value).build();
      case FilterActionType.FILTER_SQL:
      default:
        switch(action){
          case CONTAINS:
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param LIKE '%$value%'").build();
          case EQUALS:
//            num numValue = num.tryParse(value);
//            dynamic parsedValue = numValue != null ? value : "\"${value}\"";
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param = \"$value\"").build();
          case IN:
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param IN $value").build();
          case SPECIFIED:
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param IS ${value ? 'NOT NULL' : 'NULL'}").build();
          case GREATER_OR_EQUAL_THAN:
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param >= $value").build();
          case GREATER_THAN:
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param > $value").build();
          case LESS_OR_EQUAL_THAN:
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param <= $value").build();
          case LESS_THAN:
            return MapBuilder<String, dynamic>(multiKey: true).put(param, "$param < $value").build();
        }
    }
    return MapBuilder<String, dynamic>(multiKey: true).put(param, "").build();
  }

  static String parseToSQLAction(MapEntry<String, dynamic> entry){
    if(entry == null || entry.key == null) return "";
    List<String> split = entry.key.split('.');
    String param = split != null && split.length > 0 ? split[0] : null;
    String action = split != null && split.length > 1 ? split[1] : null;
    dynamic value = entry.value;

    if(action == null || action.isEmpty || param == null) return "";

    switch(action){
      case CONTAINS:
        return "$param LIKE '%$value%'";
      case EQUALS:
        return "$param = $value";
      case IN:
        return "$param IN $value";
      case SPECIFIED:
        return "$param IS ${value ? 'NOT NULL' : 'NULL'}";
      case GREATER_OR_EQUAL_THAN:
        return "$param >= $value";
      case GREATER_THAN:
        return "$param > $value";
      case LESS_OR_EQUAL_THAN:
        return "$param <= $value";
      case LESS_THAN:
        return "$param < $value";
    }
    return "";
  }
}

enum FilterActionType {
  FILTER_JPA,
  FILTER_SQL,
  FILTER_NONE,
}

 /// Run flutter packages pub run build_runner build in order to changes take effect
abstract class Filter<T, V> extends Jsonable<T>
{
  String param;
  SortMode _sortMode;
  V value;
  List<V> values;
  String paramAction;
  bool paramSpecified;
  FilterActionType filterActionType;

  @JsonKey(ignore: true) MapBuilder<String, dynamic> _filtersBuilder;

  Filter({@required this.param, SortMode sortMode, this.filterActionType})
  {
    _sortMode = sortMode;
    this.filterActionType ??= FilterActionType.FILTER_JPA;
    resetFiltersBuilder();
  }

  bool get isReady => isValid;
  bool get isValid => param != null && (sortMode != null || paramAction != null);

  void clear(){
    param = null;
    sortMode = null;
    reset();
  }

  void reset(){
    sortMode = null;
    resetValues();
  }

  void resetValues(){
    value = null;
    values = null;
    paramSpecified = null;
    paramAction = null;
    resetFiltersBuilder();
  }

  T action(V value, String action){
    resetValues();
    this.value = value;
    this.paramAction = action;
    return this as T;
  }

  T equals(V value) => action(value, FilterActions.EQUALS);

  T specified(bool value){
    resetValues();
    paramSpecified = value;
    this.paramAction = FilterActions.SPECIFIED;
    return this as T;
  }

  T isIn(List<V> values){
    resetValues();
    this.values = values;
    this.paramAction = FilterActions.IN;
    return this as T;
  }

  set sortMode(SortMode value) => _sortMode = value == SortMode.NONE ? null : value;
  SortMode get sortMode => _sortMode;

  MapBuilder<String, dynamic> get filtersBuilder => _filtersBuilder;
  void resetFiltersBuilder() => _filtersBuilder = MapBuilder(multiKey: true);

  bool get isEmpty {
    return build().isEmpty;
  }

  Map<String, dynamic> build() {
    resetFiltersBuilder();
    if(value != null) filtersBuilder.putAll(FilterActions.build(filterActionType, param, paramAction, value)); else
    if(values != null && values.isNotEmpty)
      values.forEach((value) => filtersBuilder.putAll(FilterActions.build(filterActionType, param, paramAction, value))); else
    if(paramSpecified != null) filtersBuilder.putAll(FilterActions.build(filterActionType, param, paramAction, paramSpecified));
    if(sortMode != null) filtersBuilder.putAll(SortModeUtils.build(sortMode, param, filterActionType: filterActionType));
    return filtersBuilder.build();
  }
}