import 'package:adeptarch/src/api/subscription_response.dart';
import 'package:adeptarch/src/api/subscription_response_conveter.dart';
import 'package:adeptarch/src/resource/subscription_resource.dart';
import 'package:adeptarch/src/utils/adept_publish_subject.dart';
import 'package:adeptutils/adeptutils.dart';
import 'dart:async';

import 'package:graphql_flutter/graphql_flutter.dart';

typedef SubscriptionCall = Future<Stream<FetchResult>> Function();
typedef OnSubscriptionDataSuccess<DataType> = Future<SubscriptionResource> Function({SubscriptionResource<DataType> resource, SubscriptionResponse<DataType> response});
typedef OnSubscriptionDataFailed = Future<SubscriptionResource> Function({dynamic exception, SubscriptionResponse response});
typedef OnSubscriptionErrorManagement = Future<SubscriptionResource> Function(SubscriptionResource resource, {SubscriptionResponse response});
typedef OnSubscriptionFailed = Future<SubscriptionResource> Function(dynamic exception);

class SubscriptionResourceNotifier<DataType>
{
  StreamSubscription _streamSubscription;
  AdeptPublishSubject<SubscriptionResource<DataType>> _publishSubject;
  SubscriptionCall _subscriptionCall;
  OnSubscriptionDataSuccess<DataType> _onLoadedFromAPI;
  OnSubscriptionDataFailed _onLoadFromAPIFailed;
  OnSubscriptionFailed _onFailed;
  OnSubscriptionErrorManagement _onErrorManagement;
  SubscriptionResponseConverter<DataType> _dataTypeResponseConverter;

  Future<Stream<FetchResult>> subscriptionCall() => _subscriptionCall != null ? _subscriptionCall() : null;

  Future<SubscriptionResource> onLoadedFromAPI({SubscriptionResource<DataType> resource, SubscriptionResponse<DataType> response}) async {
    if(_onLoadedFromAPI != null) resource = await _onLoadedFromAPI(resource: resource, response: response);
    else{
      if(response != null && response.isSuccessful)
        return onLoadedFromAPI(resource: SubscriptionResource.data(response.data, extras: response.extensions));
    }
    if(_publishSubject != null) _publishSubject.post(resource);
    return resource;
  }

  Future<SubscriptionResource> onLoadFromAPIFailed({dynamic exception, SubscriptionResponse response}) async
  {
    SubscriptionResource<DataType> resource;
    if(_onLoadFromAPIFailed != null) resource = await _onLoadFromAPIFailed(exception: exception, response: response);
    else{
      resource = SubscriptionResource.error(
        null,
        ExceptionUtils.parseException(exception).toString(),
        exception: exception,
        extras: response?.errors,
        code: response?.statusCode
      );
      if(_onErrorManagement != null) resource = await _onErrorManagement(resource, response: response);
    }
    _publishSubject.post(resource);
    return resource;
  }

  Future<SubscriptionResource> onFailed(dynamic exception) async
  {
    SubscriptionResource<DataType> resource;
    if(_onFailed != null) resource = await _onFailed(exception);
    else{
      resource = SubscriptionResource.error(
        null,
        exception != null ? exception.toString() : null,
        exception: exception
      );
      if(_onErrorManagement != null) resource = await _onErrorManagement(resource);
    }
    _publishSubject.post(resource);
    return resource;
  }

  SubscriptionResourceNotifier({
    bool isListResponse,
    DataType dataType,
    AdeptPublishSubject<SubscriptionResource<DataType>> publishSubject,
    SubscriptionCall createAPICall,
    OnSubscriptionDataSuccess<DataType> onLoadedFromAPI,
    OnSubscriptionDataFailed onLoadFromAPIFailed,
    OnSubscriptionFailed onFailed,
    OnSubscriptionErrorManagement onErrorManagement,
    SubscriptionResponseConverter<DataType> dataTypeResponseConverter
  })
  {
    _publishSubject = publishSubject;
    _subscriptionCall = createAPICall;
    _onLoadedFromAPI = onLoadedFromAPI;
    _onLoadFromAPIFailed = onLoadFromAPIFailed;
    _onFailed = onFailed;
    _onErrorManagement = onErrorManagement;
    _dataTypeResponseConverter = dataTypeResponseConverter ?? SubscriptionResponseConverter<DataType>(data: dataType);
    if(_publishSubject == null) _publishSubject = new AdeptPublishSubject<SubscriptionResource<DataType>>();
  }

  _onData(final FetchResult data) async {
    try {
      SubscriptionResponse<DataType> response = SubscriptionResponse.fromMapQueryResult(data, _dataTypeResponseConverter.toDataResponseConverter);

      if (response.isSuccessful)
        await onLoadedFromAPI(response: response);
      else
        await onLoadFromAPIFailed(response: response);
    } catch (e) {
      await onLoadFromAPIFailed(exception: e);
    }
  }

  _onError(final Object error) async{
    if(_onLoadFromAPIFailed != null)
      _onLoadFromAPIFailed(exception: (error is SubscriptionError) ? error.payload : error);
    else{
      SubscriptionResource<DataType> resource = SubscriptionResource.error(
          null,
          ExceptionUtils.parseException(error).toString(),
          exception: error,
          extras: error,
//          code: response?.statusCode
      );
      if(_onErrorManagement != null) resource = await _onErrorManagement(resource);
      _publishSubject.post(resource);
    }
  }

  _onDone() =>
    _publishSubject.post(SubscriptionResource.done());

  AdeptPublishSubject<SubscriptionResource<DataType>> asPublishSubject() {
    start();
    return _publishSubject;
  }

  start() async{
    try{
      _streamSubscription = (await subscriptionCall()).listen(_onData, onError: _onError, onDone: _onDone,);
    }catch(e){
      onFailed(e);
    }
  }
  AdeptPublishSubject<SubscriptionResource<DataType>> get publishSubject =>
      _publishSubject;
  bool get isPaused => _streamSubscription?.isPaused;
  pause([Future resumeSignal]) => _streamSubscription?.pause(resumeSignal);
  resume() => _streamSubscription?.resume();
  Future cancel() => _streamSubscription?.cancel();
}