import 'package:adeptarch/src/widgets/widgets.dart';
import 'package:adeptarch/src/widgets/sort_view.dart';
import 'package:adeptarch/src/filters/filters.dart';
import 'package:flutter/material.dart';

class NumberFilterView extends BaseFilterView<NumberFilter> {

  String textExceptionValueValidation;
  String textValue;
  String textValueMin;
  String textValueMax;
  String textUnspecified;
  String textEqualsTo;
  String textGreaterThan;
  String textGreaterOrEqualThan;
  String textLessThan;
  String textLessOrEqualThan;
  String textRange;

  NumberFilterView({
    FilterController<NumberFilter> filterController,
    String hint,
    bool allowSorting,
    this.textValue,
    this.textValueMin,
    this.textValueMax,
    this.textUnspecified,
    this.textEqualsTo,
    this.textGreaterThan,
    this.textGreaterOrEqualThan,
    this.textLessThan,
    this.textLessOrEqualThan,
    this.textRange,
    this.textExceptionValueValidation,
  }) : super(
    filterController: filterController,
    hint: hint,
    allowSorting: allowSorting,
  ){
    textValue ??= "Value";
    textValueMin ??= "Value min";
    textValueMax ??= "Value max";
    textUnspecified ??= "Unspecified";
    textEqualsTo ??= "Equals to";
    textGreaterThan ??= "Greater than";
    textGreaterOrEqualThan ??= "Greater or equal than";
    textLessThan ??= "Less than";
    textLessOrEqualThan ??= "Less or equal than";
    textRange ??= "Range";
    textExceptionValueValidation ??= "Value must be an integer number.";
  }

  @override
  State createState()  => _NumberFilterViewState();
}

class _NumberFilterViewState extends State<NumberFilterView> {
  Color colorPrimary, colorError, colorTextTertiary;
  final textValueMinController = AdeptTextEditingController();
  final textValueMaxController = AdeptTextEditingController();
  List<DropdownMenuItem<int>> intFilterActions;

  @override
  void initState() {
    super.initState();
    init();
    listeners();
  }

  void init() {
    onFilterChanged();
    generateActionItems();
  }

  void generateActionItems(){
    intFilterActions = [
      DropdownMenuItem(
        value: NUMBER_FILTER_ALL,
        child: Text(
          widget.textUnspecified,
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        ),
      ),
      DropdownMenuItem(
        value: NUMBER_FILTER_EQUALS,
        child: Text(
          widget.textEqualsTo,
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        ),
      ),
      DropdownMenuItem(
        value: NUMBER_FILTER_GREATER_THAN,
        child: Text(
          widget.textGreaterThan,
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        ),
      ),
      DropdownMenuItem(
        value: NUMBER_FILTER_GREATER_OR_EQUAL_THAN,
        child: Text(
          widget.textGreaterOrEqualThan,
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        ),
      ),
      DropdownMenuItem(
        value: NUMBER_FILTER_LESS_THAN,
        child: Text(
          widget.textLessThan,
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        ),
      ),
      DropdownMenuItem(
        value: NUMBER_FILTER_LESS_OR_EQUAL_THAN,
        child: Text(
          widget.textLessOrEqualThan,
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        ),
      ),
      DropdownMenuItem(
        value: NUMBER_FILTER_RANGE,
        child: Text(
          widget.textRange,
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        ),
      ),
    ];
  }

  void listeners() {
    widget.filterController.addListener(onFilterChanged);
    textValueMinController.addListener(onValueMinTextChanged);
    textValueMaxController.addListener(onValueMaxTextChanged);
  }

  void onFilterChanged(){
    NumberFilter filter = widget.filterController.value;
    if(filter != null){
      textValueMinController.text = "${filter.valueMin ?? ""}";
      textValueMaxController.text = "${filter.valueMax ?? ""}";
    }
  }

  void onValueMinTextChanged(){
    if(TextUtils.isEmpty(textValueMinController.text)){
      widget.filterController.value.valueMin = null;
      return;
    }
    int value = int.tryParse(textValueMinController.text);
    if(value != null){
      widget.filterController.value.valueMin = value;
      if(widget.filterController.error != null){
        widget.filterController.clearError();
        setState(() {});
      }
    } else{
      bool hadError = widget.filterController.error != null;
      widget.filterController.value.valueMin = null;
      if(!hadError) {
        widget.filterController.error = widget.textExceptionValueValidation;
        setState(() {});
      }
    }
  }

  void onValueMaxTextChanged(){
    if(TextUtils.isEmpty(textValueMaxController.text)){
      widget.filterController.value.valueMax = null;
      return;
    }
    int value = int.tryParse(textValueMaxController.text);
    if(value != null){
      widget.filterController.value.valueMax = value;
      if(widget.filterController.error != null){
        widget.filterController.clearError();
        setState(() {});
      }
    } else{
      bool hadError = widget.filterController.error != null;
      widget.filterController.value.valueMax = null;
      if(!hadError) {
        widget.filterController.error = widget.textExceptionValueValidation;
        setState(() {});
      }
    }
  }

  void checkColors(context){
    colorPrimary ??= Theme.of(context).primaryColor;
    colorError ??= Theme.of(context).errorColor;
    colorTextTertiary ??= Colors.grey.shade600;
  }

  @override
  Widget build(BuildContext context) {
    checkColors(context);
    return Container(
      child: Column(
        mainAxisAlignment: MainAxisAlignment.start,
        crossAxisAlignment: CrossAxisAlignment.start,
        children: <Widget>[
          Text(
            widget.hint,
            style: TextStyle(color: widget.filterController.value.paramAction != null ? colorPrimary : colorTextTertiary, fontSize: 12),
          ),
          widget.filterController.error == null ? Container() :
          Container(
            child: Text(
              widget.filterController.error,
              style: TextStyle(color: colorError, fontSize: 12),
            ),
          ),
          Row(
            children: <Widget>[
              Expanded(
                child: Theme(
                  data: ThemeData(brightness: Theme.of(context).brightness ?? Brightness.light),
                  child: DropdownButton<int>(
                    isExpanded: widget.filterController.value.paramAction == null,
                    items: intFilterActions,
                    value: widget.filterController.value.numberFilterType,
                    onChanged: (numberFilterType){
                      if(numberFilterType == null) {
                        widget.filterController.reset();
                      } else{
                        int value = int.tryParse(textValueMinController.text);
                        widget.filterController.value.numberFilterType = numberFilterType;
                        widget.filterController.value.valueMin = value;
                        widget.filterController.value.valueMax = value;
                      }
                      setState(() {});
                    },
                  ),
                ),
              ),
              Visibility(
                child: Container(
                  margin: EdgeInsets.only(top: 16),
                  alignment: Alignment.topCenter,
                  child: SortView(
                    sortMode: widget.filterController.value.sortMode,
                    listener: (oldMode, newMode){
                      widget.filterController.value.sortMode = newMode;
                    },
                  ),
                ),
                visible: widget.allowSorting,
              )
            ],
          ),
          Row(
            children: <Widget>[
              Visibility(
                child: Expanded(
                  child: Container(
                    child: TextField(
                      controller: textValueMinController,
                      textInputAction: TextInputAction.done,
                      keyboardType: TextInputType.number,
                      decoration: InputDecoration(
                        labelText: widget.filterController.value.numberFilterType != NUMBER_FILTER_RANGE ? widget.textValue : widget.textValueMin,
                        contentPadding: EdgeInsets.all(8),
                        focusedBorder: OutlineInputBorder(
                          borderSide: BorderSide(
                              width: 1,
                              color: Colors.grey.shade300
                          ),
                        ),
                        enabledBorder: OutlineInputBorder(
                          borderSide: BorderSide(
                              width: 1,
                              color: Colors.grey.shade300
                          ),
                        ),
                      ),
                    ),
                  ),
                ),
                visible: widget.filterController.value.numberFilterType != NUMBER_FILTER_ALL,
              ),
              Visibility(
                child: Text(" --- "),
                visible: widget.filterController.value.numberFilterType == NUMBER_FILTER_RANGE,
              ),
              Visibility(
                child: Expanded(
                  child: Container(
                    child: TextField(
                      controller: textValueMaxController,
                      textInputAction: TextInputAction.done,
                      keyboardType: TextInputType.number,
                      decoration: InputDecoration(
                        labelText: widget.textValueMax,
                        contentPadding: EdgeInsets.all(8),
                        focusedBorder: OutlineInputBorder(
                          borderSide: BorderSide(
                              width: 1,
                              color: Colors.grey.shade300
                          ),
                        ),
                        enabledBorder: OutlineInputBorder(
                          borderSide: BorderSide(
                              width: 1,
                              color: Colors.grey.shade300
                          ),
                        ),
                      ),
                    ),
                  ),
                ),
                visible: widget.filterController.value.numberFilterType == NUMBER_FILTER_RANGE,
              ),
            ],
          ),
        ],
      ),
    );
  }

  @override
  void dispose() {
    widget.filterController.removeListener(onFilterChanged);
    super.dispose();
  }

}